from web3 import AsyncWeb3
from web3.contract import AsyncContract
from web3.middleware import ExtraDataToPOAMiddleware

from . import config
from .data_types import Chain, Token
from .client import client


async def make_w3(chain: Chain) -> AsyncWeb3:
    w3 = AsyncWeb3(AsyncWeb3.AsyncHTTPProvider(chain.rpc_url))
    w3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)

    assert await w3.is_connected()

    return w3


def make_token_contract(w3: AsyncWeb3, token: Token) -> AsyncContract:
    return w3.eth.contract(
        address=AsyncWeb3.to_checksum_address(token.contract_address),
        abi=config.erc20_abi,
    )


def make_order_book_contract(w3: AsyncWeb3, token: Token) -> AsyncContract:
    return w3.eth.contract(
        address=client.deployments[token.chain.id]["contracts"]["order_book"],
        abi=config.order_book_abi,
    )



# Note: this module should really be merged with the data_types module, but is kept separate to avoid circular imports in the client module

from __future__ import annotations
from enum import IntEnum

from eth_typing import ChecksumAddress
from hexbytes import HexBytes


class ChainId(IntEnum):
    ETHEREUM = 1
    OP = 10
    BSC = 56
    POLYGON = 137
    OPBNB = 204
    MANTLE = 5000
    BASE = 8453
    MODE = 34443
    ARBITRUM = 42161
    CELO = 42220
    AVALANCHE_C_CHAIN = 43114
    BLAST = 81457
    SCROLL = 54352

    @staticmethod
    def from_name(name: str) -> ChainId:
        return CHAIN_IDS[name]

    def __str__(self):
        return CHAIN_NAMES[self]


CHAIN_NAMES = {
    ChainId.ETHEREUM: "ethereum",
    ChainId.OP: "optimism",
    ChainId.BSC: "bsc",
    ChainId.POLYGON: "polygon",
    ChainId.OPBNB: "opbnb",
    ChainId.MANTLE: "mantle",
    ChainId.BASE: "base",
    ChainId.MODE: "mode",
    ChainId.ARBITRUM: "arbitrum",
    ChainId.CELO: "celo",
    ChainId.AVALANCHE_C_CHAIN: "avalanche",
    ChainId.BLAST: "blast",
    ChainId.SCROLL: "scroll",
}

CHAIN_IDS = {name: id for id, name in CHAIN_NAMES.items()}

SCANNERS = {
    ChainId.ETHEREUM: "https://etherscan.io",
    ChainId.OP: "https://optimistic.etherscan.io",
    ChainId.BSC: "https://bscscan.com",
    ChainId.POLYGON: "https://polygonscan.com",
    ChainId.OPBNB: "https://opbnbscan.com",
    ChainId.MANTLE: "https://explorer.mantle.xyz",
    ChainId.BASE: "https://basescan.org",
    ChainId.MODE: "https://modescan.io",
    ChainId.ARBITRUM: "https://arbiscan.io",
    ChainId.CELO: "https://explorer.celo.org/mainnet",
    ChainId.AVALANCHE_C_CHAIN: "https://snowscan.xyz",
    ChainId.BLAST: "https://blastscan.io",
    ChainId.SCROLL: "https://scrollscan.com",
}


class Scanner:
    @staticmethod
    def address(chain_id: ChainId, wallet: ChecksumAddress) -> str:
        return f"{SCANNERS[chain_id]}/address/{wallet}"

    @staticmethod
    def transaction(chain_id: ChainId, transaction_hash: HexBytes) -> str:
        return f"{SCANNERS[chain_id]}/tx/{transaction_hash.to_0x_hex()}"

    @staticmethod
    def token(token: Token) -> str:
        return f"{SCANNERS[token.chain.id]}/token/{token.contract_address}"



from eth_typing import ChecksumAddress
from typing import Any

from .client import client
from .constants import ChainId
from . import config


class Token:
    __slots__ = ("chain", "symbol", "chain_id", "contract_address", "decimals")

    def __init__(self, chain_id: ChainId, symbol: str):
        self.symbol = symbol
        self.chain = Chain(chain_id)
        asset_data = self.chain.data["assets"][self.symbol]
        self.contract_address: ChecksumAddress = asset_data["address"]
        self.decimals: int = asset_data["decimals"]

    @classmethod
    def from_str(cls, identifier: str):
        chain_name, symbol = identifier.split("-")
        chain_id = ChainId.from_name(chain_name)
        return cls(chain_id, symbol)

    @classmethod
    def from_contract_address(cls, chain: ChainId, contract_address: ChecksumAddress):
        symbol = client.symbol_by_contract[(chain, contract_address)]
        return cls(chain, symbol)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Token)
            and self.chain == other.chain
            and self.symbol == other.symbol
        )

    def __repr__(self) -> str:
        return f"{self.chain}-{self.symbol}"

    def is_stablecoin(self) -> bool:
        upper = self.symbol.upper()

        return any(
            map(lambda symbol: symbol in upper, ("USD", "EUR", "JPY", "GPB", "CHF"))
        ) or upper in ("FRAX", "DAI", "MIM")


class Chain:
    __slots__ = ("id", "data", "lz_cid", "rpc_url")

    def __init__(self, id: ChainId):
        self.id = id
        self.data = client.deployments[self.id]
        self.lz_cid: int = self.data["lz_cid"]  # type: ignore
        self.rpc_url = config.rpc_urls[self.id]

    @classmethod
    def from_id(cls, id: int):
        return cls(ChainId(id))

    @classmethod
    def from_name(cls, name: str):
        return cls(ChainId.from_name(name))

    @property
    def name(self) -> str:
        return str(self.id)

    def __eq__(self, other) -> bool:
        return isinstance(other, Chain) and self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    def __repr__(self) -> str:
        return self.name



from importlib import resources
from importlib.abc import Traversable
import json
import os
from typing import Any

from .constants import ChainId

# Anything on these chains will be invisible to the client
excluded_chains = frozenset(
    (
        ChainId.ETHEREUM,
        ChainId.MODE,
        ChainId.CELO,
        ChainId.MANTLE,
        ChainId.OPBNB,
        ChainId.SCROLL,
        ChainId.BLAST,
    )
)

rpc_urls = {
    ChainId.ETHEREUM: "https://long-capable-tree.quiknode.pro/4f8abee71b92694624d5f6a5aac43bf273b76db1/",
    ChainId.OP: "https://alien-dry-lake.optimism.quiknode.pro/9e3364e544a78fa0581658f542d58d8c02cd13ba/",
    ChainId.BSC: "https://necessary-thrumming-diamond.bsc.quiknode.pro/f2f3a08170d298c0bd742d77ea0462afcec15dd4/",
    ChainId.POLYGON: "https://holy-restless-vineyard.matic.quiknode.pro/559b966d2ea5ab37e95abee0cf1049a10971e30d/",
    # ChainId.OPBNB: "Only has regular BSC?",
    ChainId.MANTLE: "https://still-compatible-wind.mantle-mainnet.quiknode.pro/277b01e4afb73df889a08ba018dc14e354f5f0da/",
    ChainId.BASE: "https://polished-spring-star.base-mainnet.quiknode.pro/19455fd433fb2639609315f8588c3a58a5a9a10f/",
    # ChainId.MODE: "Not yet available on Quicknode?",
    ChainId.ARBITRUM: "https://prettiest-patient-scion.arbitrum-mainnet.quiknode.pro/2d53fa7ffc71e31afb3113f96c54519fcd6516e2/",
    ChainId.CELO: "https://attentive-alien-water.celo-mainnet.quiknode.pro/9c1b86178a6c7ebc5c772663ac82b51003ef8a81/",
    ChainId.AVALANCHE_C_CHAIN: "https://withered-soft-field.avalanche-mainnet.quiknode.pro/f0478bf89e96d35ee8351213a1120fe4ba292849/ext/bc/C/rpc/",
    ChainId.BLAST: "https://warmhearted-few-scion.blast-mainnet.quiknode.pro/22d44d1f9a1a57adf90c0efd2d644977c3dd5a23/",
    ChainId.SCROLL: "https://sleek-crimson-flower.scroll-mainnet.quiknode.pro/37e22c7823176a82f7cea7b89b5c37786b76a810/",
    # tron: "https://practical-restless-meme.tron-mainnet.quiknode.pro/6088a1547d6e57bd2b9376aebca00366523ae405/jsonrpc",
    # solana: "https://cosmological-soft-dust.solana-mainnet.quiknode.pro/c4fc0753cc3e5219724fcbf042ce9ce0abd84590/",
}

backend_url = os.environ.get(
    "MACH_BACKEND_URL", "https://cache-half-full-production.fly.dev"
)

endpoints = {
    "orders": "/v1/orders",
    "gas": "/v1/orders/gas",
    "quotes": "/v1/quotes",
    "points": "/v1/points",
    "token_balances": "/tokenBalances",
    "get_config": "/get_config",
}


def load_abi(path: Traversable) -> Any:
    with path.open("r") as abi:
        return json.load(abi)


# Relative to the root of the repository
abi_path = resources.files("abi")

order_book_abi = load_abi(abi_path / "mach" / "order_book.json")

erc20_abi = load_abi(abi_path / "ethereum" / "erc20.json")



from __future__ import annotations
import logging
from pprint import pformat
import typing
from typing import Any, Iterator, NamedTuple, TypedDict

from aiohttp import ClientSession, ClientResponse
from eth_typing import ChecksumAddress
from hexbytes import HexBytes

from . import config
from .constants import ChainId


# Note: we should really just be deserializing the backend types with BaseModel.validate_from_json
# - https://github.com/tristeroresearch/cache-half-full/blob/62b31212f0456e4fad564021289816d39345b49b/backend/api/v1/endpoints/quotes.py#L51
# - https://docs.pydantic.dev/latest/api/base_model/#pydantic.BaseModel.model_validate_json


class Order(TypedDict):
    id: str
    taker_address: ChecksumAddress
    maker_address: ChecksumAddress
    src_asset_address: ChecksumAddress
    dst_asset_address: ChecksumAddress
    src_chain: str
    dst_chain: str
    src_amount: int
    dst_amount: int
    created_at: str
    filled_at: str
    expires_at: str
    completed: bool
    place_tx: HexBytes
    fill_tx: HexBytes
    eta: int


class GasEstimate(TypedDict):
    gas_estimate: int
    gas_price: int


class Quote(TypedDict):
    wallet_address: str
    src_chain: str
    dst_chain: str
    src_amount: int
    dst_amount: int
    bond_amount: int
    bond_fee: int
    src_asset_address: ChecksumAddress
    bond_asset_address: ChecksumAddress
    challenge_offset: int
    challenge_window: int
    invalid_amount: bool
    liquidity_source: str
    created_at: str
    expires_at: str


class WalletPoints(NamedTuple):
    wallet: ChecksumAddress
    points: int


# This class is made to be used as a singleton (see below)
class MachClient:
    backend_url = config.backend_url

    # Routes
    orders = config.backend_url + config.endpoints["orders"]
    gas = config.backend_url + config.endpoints["gas"]
    quotes = config.backend_url + config.endpoints["quotes"]
    points = config.backend_url + config.endpoints["points"]
    token_balances = config.backend_url + config.endpoints["token_balances"]
    get_config = config.backend_url + config.endpoints["get_config"]

    # Make some configuration accessible to the user
    excluded_chains = config.excluded_chains
    rpc_urls = config.rpc_urls

    def __init__(self):
        self.logger = logging.getLogger("mach-client")

        # Contains data for all chains
        self.deployments: dict[ChainId, dict[str, Any]] = {}

        # Contains only data for supported chains
        self.gas_tokens: dict[ChainId, str] = {}
        self.chains: set[ChainId] = set()
        self.symbol_by_contract: dict[tuple[ChainId, ChecksumAddress], str] = {}

    _singleton = None

    @classmethod
    async def create(cls):
        if MachClient._singleton:
            return MachClient._singleton

        MachClient._singleton = cls()
        await MachClient._singleton.init()
        return MachClient._singleton

    async def init(self) -> None:
        self.session = ClientSession()
        self.session.headers.update(
            {
                "accept": "application/json",
                "Content-Type": "application/json",
            }
        )

        # Fetch config from API
        async with self.session.get(self.get_config) as response:
            self.raw_deployments: dict[str, dict[str, Any]] = (
                await self._validate_response(response)
            )["deployments"]

        # Initialize attributes
        for chain_name, chain_data in self.raw_deployments.items():
            chain_id = ChainId.from_name(chain_name)

            if chain_id in self.excluded_chains:
                continue

            if chain_id not in self.rpc_urls:
                self.logger.warning(f"{chain_id} RPC URL missing from config")
                continue

            # Convert from chain names to chain IDs
            del chain_data["chain_id"]
            self.deployments[chain_id] = chain_data

            self.symbol_by_contract.update(
                {
                    (chain_id, token_data["address"]): symbol
                    for symbol, token_data in chain_data["assets"].items()
                }
            )

            self.chains.add(chain_id)

            # The gas token has "wrapped": True
            gas_tokens: Iterator = filter(
                lambda item: item[1].get("wrapped"), chain_data["assets"].items()
            )

            # Maps (symbol name, symbol data) -> symbol name
            gas_tokens = map(lambda item: item[0], gas_tokens)

            try:
                gas_token = next(gas_tokens)
            except StopIteration as e:
                self.logger.warning(f"No gas token in config of {chain_id}")
                continue

            self.gas_tokens[chain_id] = gas_token

            assert not (
                gas_token_2 := next(gas_tokens, None)
            ), f"Multiple gas tokens on {chain_id}: {gas_token}, {gas_token_2}"

    async def close(self) -> None:
        if hasattr(self, "session"):
            await self.session.close()

    async def _validate_response(self, response: ClientResponse) -> Any:
        match response.status:
            case 200:
                return await response.json()

            case 422:
                raise ValueError(
                    f"Validation error - invalid request: {pformat(await response.json())}"
                )

            case _:
                raise RuntimeError(
                    f"Unknown status code {response.status}: {pformat(await response.json())}"
                )

    def _format_order(self, order: Order) -> Order:
        order["place_tx"] = HexBytes(order["place_tx"])

        if fill_tx := order.get("fill_tx", None):
            order["fill_tx"] = HexBytes(fill_tx)

        return order

    async def submit_order(self, chain: ChainId, place_taker_tx: HexBytes) -> Order:
        json = {
            "chain": str(chain),
            "place_taker_tx": place_taker_tx.hex(),
        }

        async with self.session.post(self.orders, json=json) as response:
            return self._format_order(await self._validate_response(response))

    async def get_orders(self, wallet: ChecksumAddress) -> list[Order]:
        params = {
            "wallet": wallet,
        }

        async with self.session.get(self.orders, params=params) as response:
            orders = await self._validate_response(response)

        return list(map(self._format_order, orders))

    async def estimate_gas(self, chain: ChainId) -> GasEstimate:
        params = {"chain": str(chain)}

        async with self.session.get(self.gas, params=params) as response:
            json = await self._validate_response(response)

        return typing.cast(GasEstimate, json)

    async def request_quote(
        self,
        src_token: Token,  # type: ignore
        dest_token: Token,  # type: ignore
        src_amount: int,
        wallet: ChecksumAddress,
    ) -> Quote:
        json = {
            "dst_asset_address": dest_token.contract_address,
            "dst_chain": dest_token.chain.name,
            "src_amount": src_amount,
            "src_asset_address": src_token.contract_address,
            "src_chain": src_token.chain.name,
            "wallet_address": wallet,
        }

        async with self.session.post(self.quotes, json=json) as response:
            json = await self._validate_response(response)

        return typing.cast(Quote, json)

    async def get_all_points(self, limit: int = 10) -> list[WalletPoints]:
        params = {
            "limit": limit,
        }

        async with self.session.get(self.points, params=params) as response:
            json = await self._validate_response(response)

        return list(
            map(
                lambda item: WalletPoints(item["wallet"], int(item["points"])),
                json,
            )
        )

    async def get_points(self, wallet: ChecksumAddress) -> int:
        async with self.session.get(self.points + f"/{wallet}") as response:
            json = await self._validate_response(response)

        return int(json["points"])

    async def get_token_balances(
        self, wallet_address: ChecksumAddress
    ) -> dict[ChainId, dict[str, int]]:
        params = {"wallet_address": wallet_address}

        async with self.session.get(self.token_balances, params=params) as response:
            raw_balances: dict[str, dict[str, int]] = (
                await self._validate_response(response)
            )["balances"]

        return dict(
            # Filter out chains we don't support
            filter(
                lambda item: item[0] in self.chains,
                # Map chain names to IDs
                map(
                    lambda item: (ChainId.from_name(item[0]), item[1]),
                    raw_balances.items(),
                ),
            )
        )


# Singleton - call await client.init() from your application
client = MachClient()



import logging as _logging

from .client import GasEstimate, MachClient, Order, Quote, WalletPoints, client
from .constants import ChainId, Scanner
from .data_types import Chain, Token
from .log import LogContextAdapter, Logger, make_logger as _make_logger

__all__ = [
    "Chain",
    "ChainId",
    "GasEstimate",
    "LogContextAdapter",
    "Logger",
    "MachClient",
    "Order",
    "Quote",
    "Scanner",
    "Token",
    "WalletPoints",
    "client",
]

_make_logger("mach-client", _logging.INFO)
